/* WCETargetDataLine.java
   Copyright (C) 2006, 2007, 2008, 2009 Mysaifu.com

This file is part of Mysaifu JVM.

Mysaifu JVM is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; the version 2.

Mysaifu JVM is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Classpath; see the file COPYING.  If not, write to the
Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301 USA.

Linking this library statically or dynamically with other modules is
making a combined work based on this library.  Thus, the terms and
conditions of the GNU General Public License cover the whole
combination.

As a special exception, the copyright holders of this library give you
permission to link this library with independent modules to produce an
executable, regardless of the license terms of these independent
modules, and to copy and distribute the resulting executable under
terms of your choice, provided that you also meet, for each linked
independent module, the terms and conditions of the license of that
module.  An independent module is a module which is not derived from
or based on this library.  If you modify this library, you may extend
this exception to your version of the library, but you are not
obligated to do so.  If you do not wish to do so, delete this
exception statement from your version. */


package gnu.javax.sound.sampled.wce;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioPermission;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Control;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.TargetDataLine;

/**
 * ̓C
 */
public class WCETargetDataLine
    extends WCEDataLine
    implements TargetDataLine
{
  /**
   * lCeBuf[^
   */
  private int nativeData;

  /**
   * o̓C̃CX^X쐬B
   * 
   * @param info Linȅ
   */
  public WCETargetDataLine(DataLine.Info info)
  {
    super(info);
  }

  /**
   * ̃CftHgݒŃI[v
   */
  public void open() throws LineUnavailableException, IllegalArgumentException,
      SecurityException
  {
    open(getFormat());
  }

  /**
   * ̃CI[v
   * 
   * @param format tH[}bg
   * @param bufferSize obt@TCY
   */
  public synchronized void open(AudioFormat format, int bufferSize)
      throws LineUnavailableException, IllegalArgumentException,
      IllegalStateException, SecurityException
  {
    if (isOpen())
      {
        throw new IllegalStateException("Already open");
      }

    // ZLeB`FbN
    SecurityManager sm = System.getSecurityManager();
    if (sm != null)
      {
        // "record" `FbN
        sm.checkPermission(new AudioPermission("record"));
      }

    // tH[}bg`FbN
    // T|[gĂ͈̂ȉ̂Q
    // o 8rbgȂPCM
    // o 16rbgtPCM
    boolean validFormat = false;
    int sampleSizeInBits = format.getSampleSizeInBits();
    AudioFormat.Encoding encoding = format.getEncoding();
    switch (sampleSizeInBits)
      {
      case 8:
        // Wrbg̏ꍇ͕Ȃ
        if (AudioFormat.Encoding.PCM_UNSIGNED.equals(encoding))
          {
            validFormat = true;
          }
        break;

      case 16:
        // 16rbg̏ꍇ͕t
        if (AudioFormat.Encoding.PCM_SIGNED.equals(encoding))
          {
            validFormat = true;
          }
        break;
      }
    if (format.getChannels() > 0)
      {
        // `l`FbN
        validFormat = true;
      }

    if (! validFormat)
      {
        throw new LineUnavailableException("Unsupported format:" + format);
      }

    super.open();
    this.nativeData = openNative(bufferSize, format.getSampleRate(),
                                 format.getSampleSizeInBits(),
                                 format.getChannels(), format.getFrameSize(),
                                 format.getFrameRate(), format.isBigEndian());
    setFormat(format);
  }

  /**
   * ̃CI[v
   * 
   * @param format tH[}bg
   */
  public void open(AudioFormat format) throws LineUnavailableException,
      IllegalArgumentException, IllegalStateException, SecurityException
  {
    open(format, 0);
  }

  /**
   * Windows CẺ̓foCXI[vB
   * 
   * @param bufferSize obt@TCY
   * @param sampleRate Tv[g
   * @param sampleSizeInBits rbgPʂŕ\Tvf[^TCYB
   * @param channels `lB
   * @param frameSize 1t[̃f[^oCg
   * @param frameRate 1b̍Đt[
   * @param bigEndian f[^rbOGfBȀꍇtrueB
   * @return lCeBuf[^Bi\̂̃|C^j
   * @throws LineUnavailableException o̓foCXI[vłȂꍇB
   */
  private native int openNative(int bufferSize, float sampleRate,
                                int sampleSizeInBits, int channels,
                                int frameSize, float frameRate,
                                boolean bigEndian)
      throws LineUnavailableException, IllegalArgumentException,
      IllegalStateException, SecurityException;

  /**
   * I[fBf[^ǂݍ
   */
  public int read(byte[] b, int off, int len) throws IllegalArgumentException,
      ArrayIndexOutOfBoundsException
  {
    if (! isOpen())
      {
        // return 0;
        // AudioInputStreamread()Ƌł悤ɁA
        // -1 Ԃ
        return - 1;
      }
    return readNative(this.nativeData, b, off, len);
  }

  /**
   * I[fBIf[^ǂݍ
   * 
   * @param nativeData lCeBuf[^
   * @param b obt@
   * @param off
   * @param len
   */
  private native int readNative(int nativeData, byte[] b, int off, int len)
      throws IllegalArgumentException, ArrayIndexOutOfBoundsException;

  /**
   * CN[Y
   */
  public void close()
  {
    if (isOpen())
      {
        super.close();
        stop();
        closeNative(this.nativeData);
      }
  }

  /**
   * Windows CẺ̓foCXN[Y
   */
  private native void closeNative(int nativeData);

  public void drain()
  {
    if (! isOpen())
      {
        return;
      }
    drainNative(this.nativeData);
  }

  /**
   * obt@̃f[^ɂȂ܂ŃubN
   */
  private native void drainNative(int nativeData);

  public void flush()
  {
    if (! isOpen())
      {
        return;
      }
    flushNative(this.nativeData);
  }

  /**
   * obt@̃f[^j
   */
  private native void flushNative(int nativeData);

  public boolean isActive()
  {
    return isNativeActive(this.nativeData);
  }

  /**
   * Windows CEfoCXŉ͒ԂB
   */
  private native boolean isNativeActive(int nativeData);

  public int getBufferSize()
  {
    if (! isOpen())
      {
        return 0;
      }
    return getNativeBufferSize(this.nativeData);
  }

  /**
   * lCeBuobt@TCYԂ
   */
  private native int getNativeBufferSize(int nativeData);

  public int available()
  {
    if (! isOpen())
      {
        return 0;
      }
    return availableNative(this.nativeData);
  }

  /**
   * lCeBuCuŃubNɓǂݍ߂oCg ԂB
   */
  private native int availableNative(int nativeData);

  public int getFramePosition()
  {
    return (int) getLongFramePosition();
  }

  public long getLongFramePosition()
  {
    if (! isOpen())
      {
        return 0L;
      }
    // ToDo: 
    return AudioSystem.NOT_SPECIFIED;
  }

  public long getMicrosecondPosition()
  {
    if (! isOpen())
      {
        return 0L;
      }
    // ToDo: 
    return AudioSystem.NOT_SPECIFIED;
  }

  public float getLevel()
  {
    // ToDo: 
    return AudioSystem.NOT_SPECIFIED;
  }

  public Control[] getControls()
  {
    // ToDo: 
    return new Control[0];
  }

  public boolean isControlSupported(Control.Type control)
  {
    // ToDo: 
    return false;
  }

  public Control getControl(Control.Type control)
  {
    // ToDo: 
    throw new IllegalArgumentException("Not supported:" + control);
  }

  public void start()
  {
    if (! isOpen())
      {
        return;
      }
    super.start();
    startNative(this.nativeData);
  }

  private native void startNative(int nativeData);

  public void stop()
  {
    if (! isOpen())
      {
        return;
      }
    stopNative(this.nativeData);
    super.stop();
  }

  private native void stopNative(int nativeData);
}
